"""Tests for Nexia Home."""

import asyncio
import json
import os
from os.path import dirname
from pathlib import Path

import aiohttp
import pytest
from aioresponses import aioresponses
from aioresponses.compat import URL

from nexia.home import (
    LoginFailedException,
    NexiaHome,
    _extract_devices_from_houses_json,
    extract_children_from_devices_json,
)
from nexia.thermostat import NexiaThermostat, clamp_to_predefined_values
from nexia.util import find_dict_with_keyvalue_in_json

pytestmark = pytest.mark.asyncio


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest.fixture(name="aiohttp_session")
async def aiohttp_session():
    """Mock session."""
    session = aiohttp.ClientSession()
    yield session
    await session.close()


def _load_fixture(filename):
    """Load a fixture."""
    test_dir = dirname(__file__)

    path = os.path.join(test_dir, "fixtures", filename)
    with open(path) as fptr:
        return fptr.read()


async def load_fixture(filename):
    """Load a fixture."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _load_fixture, filename)


async def test_login(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test login sequence."""
    persist_file = Path("nexia_config_test.conf")
    nexia = NexiaHome(aiohttp_session, state_file=persist_file)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/accounts/sign_in",
        status=206,
        body="no good",
    )
    with pytest.raises(
        ValueError,
        match="Failed to login\nno good",
    ):
        await nexia.login()

    forgot_password_url = "https://www.mynexia.com/account/forgotten_credentials"
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/accounts/sign_in",
        status=307,
        headers={aiohttp.hdrs.LOCATION: forgot_password_url},
    )
    mock_aioresponse.get(
        forgot_password_url,
    )
    with pytest.raises(
        LoginFailedException,
        match=f"Failed to login, getting redirected to {forgot_password_url}"
        f". Try to login manually on the website.",
    ):
        await nexia.login()

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/accounts/sign_in",
        payload={
            "success": True,
            "error": None,
            "result": {
                "mobile_id": 5400000,
                "api_key": "10654c0be00000000000000000000000",
                "setup_step": "done",
                "locale": "en_us",
            },
        },
    )
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/session",
        body=await load_fixture("mobile_session.json"),
    )
    await nexia.login()

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/houses/2582941",
        status=304,
    )
    assert await nexia.update() is None

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/houses/2582941",
        status=208,
        body="failing text",
    )
    with pytest.raises(
        ValueError,
        match="Unexpected http status while fetching house JSON\nfailing text",
    ):
        await nexia.update()

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/houses/2582941",
    )
    with pytest.raises(
        ValueError,
        match="Nothing in the JSON",
    ):
        await nexia.update()

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/houses/2582941",
        body=await load_fixture("mobile_houses_123456.json"),
    )
    assert await nexia.update() is not None

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/phones",
        body=await load_fixture("mobile_phones_response.json"),
    )
    assert await nexia.get_phone_ids() == [5488863]

    assert nexia.get_thermostat_ids() == [2059661, 2059676, 2293892, 2059652]
    thermostat = nexia.get_thermostat_by_id(2059661)
    assert thermostat.get_zone_ids() == [83261002, 83261005, 83261008, 83261011]
    zone = thermostat.get_zone_by_id(83261002)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/83261002/setpoints",
        body=await load_fixture("zone_response.json"),
    )
    await zone.set_heat_cool_temp(69.0, 78.0)

    assert persist_file.exists() is True
    persist_file.unlink()
    assert persist_file.exists() is False


async def test_update(aiohttp_session: aiohttp.ClientSession) -> None:
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)
    thermostat = nexia.get_thermostat_by_id(2059661)
    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83261002, 83261005, 83261008, 83261011]
    nexia.update_from_json(devices_json)
    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83261002, 83261005, 83261008, 83261011]
    nexia.update_from_json(devices_json)


async def test_idle_thermo(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get methods for an idle thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(2059661)

    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.1"
    assert thermostat.get_dev_build_number() == "1581321824"
    assert thermostat.get_device_id() == "000000"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Downstairs East Wing"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.get_variable_fan_speed_limits() == (0.35, 1.0)
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode(), "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_outdoor_temperature() == 88.0
    assert thermostat.get_relative_humidity() == 0.36
    assert thermostat.get_current_compressor_speed() == 0.0
    assert thermostat.get_requested_compressor_speed() == 0.0
    assert thermostat.get_fan_speed_setpoint() == 0.35
    assert thermostat.get_dehumidify_setpoint() == 0.50
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.get_air_cleaner_mode() == "auto"
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83261002, 83261005, 83261008, 83261011]


async def test_idle_thermo_issue_33758(mock_aioresponse: aioresponses) -> None:
    """Get methods for an idle thermostat."""
    aiohttp_session = aiohttp.ClientSession()
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33758.json"))
    nexia.update_from_json(devices_json)

    thermostat: NexiaThermostat = nexia.get_thermostat_by_id(12345678)

    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.1"
    assert thermostat.get_dev_build_number() == "1581321824"
    assert thermostat.get_device_id() == "xxxxxx"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Thermostat"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.get_variable_fan_speed_limits() == (0.35, 1.0)
    assert thermostat.get_unit() == "F"
    assert thermostat.has_humidify_support() is True
    assert thermostat.get_humidity_setpoint_limits() == (0.10, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_outdoor_temperature() == 55.0
    assert thermostat.get_relative_humidity() == 0.43
    assert thermostat.get_current_compressor_speed() == 0.0
    assert thermostat.get_requested_compressor_speed() == 0.0
    assert thermostat.get_fan_speed_setpoint() == 1
    assert thermostat.get_dehumidify_setpoint() == 0.55
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is True
    assert thermostat.has_emergency_heat() is True
    assert thermostat.is_emergency_heat_active() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.get_air_cleaner_mode() == "auto"
    assert thermostat.is_blower_active() is False

    devices = _extract_devices_from_houses_json(devices_json)
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/emergency_heat",
        payload={"result": devices[0]},
    )
    await thermostat.set_emergency_heat(True)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/emergency_heat",
        payload={"result": devices[0]},
    )
    await thermostat.set_emergency_heat(False)

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [12345678]
    await aiohttp_session.close()


async def test_idle_thermo_issue_33968_thermostat_1690380(
    aiohttp_session: aiohttp.ClientSession,
) -> None:
    """Get methods for a cooling thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33968.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [1690380]

    thermostat = nexia.get_thermostat_by_id(1690380)

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83037337, 83037340, 83037343]

    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.1"
    assert thermostat.get_dev_build_number() == "1581321824"
    assert thermostat.get_device_id() == "removed"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Thermostat"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.get_variable_fan_speed_limits() == (0.35, 1.0)
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_outdoor_temperature() == 80.0
    assert thermostat.get_relative_humidity() == 0.55
    assert thermostat.get_current_compressor_speed() == 0.41
    assert thermostat.get_requested_compressor_speed() == 0.41
    assert thermostat.get_fan_speed_setpoint() == 0.5
    assert thermostat.get_dehumidify_setpoint() == 0.55
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is True
    assert thermostat.is_emergency_heat_active() is False
    assert thermostat.get_system_status() == "Cooling"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.get_air_cleaner_mode() == "auto"
    assert thermostat.is_blower_active() is True


async def test_active_thermo(aiohttp_session):
    """Get methods for an active thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(2293892)

    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.1"
    assert thermostat.get_dev_build_number() == "1581321824"
    assert thermostat.get_device_id() == "0281B02C"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Master Suite"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.get_variable_fan_speed_limits() == (0.35, 1.0)
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_outdoor_temperature() == 87.0
    assert thermostat.get_relative_humidity() is None
    assert thermostat.get_current_compressor_speed() == 0.69
    assert thermostat.get_requested_compressor_speed() == 0.69
    assert thermostat.get_fan_speed_setpoint() == 0.35
    assert thermostat.get_dehumidify_setpoint() == 0.45
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "Cooling"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.get_air_cleaner_mode() == "auto"
    assert thermostat.is_blower_active() is True

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83394133, 83394130, 83394136, 83394127, 83394139]


@pytest.mark.skip(reason="not yet supported")
async def test_xl624(aiohttp_session):
    """Get methods for an xl624 thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_xl624.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2222222, 3333333]
    thermostat = nexia.get_thermostat_by_id(1111111)

    assert thermostat.get_model() is None
    assert thermostat.get_firmware() == "2.8"
    assert thermostat.get_dev_build_number() == "0603340208"
    assert thermostat.get_device_id() is None
    assert thermostat.get_type() is None
    assert thermostat.get_name() == "Downstairs Hall"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is False
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.10, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Cycler"]
    assert thermostat.get_current_compressor_speed() == 0.0
    assert thermostat.get_requested_compressor_speed() == 0.0
    assert thermostat.has_dehumidify_support() is False
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is False
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [12345678]


async def test_xl824_1(aiohttp_session):
    """Get methods for an xl824 thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_xl624.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2222222, 3333333]
    thermostat = nexia.get_thermostat_by_id(2222222)

    assert thermostat.get_model() == "XL824"
    assert thermostat.get_firmware() == "5.9.1"
    assert thermostat.get_dev_build_number() == "1581314625"
    assert thermostat.get_device_id() == "0167CA48"
    assert thermostat.get_type() == "XL824"
    assert thermostat.get_name() == "Family Room"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Circulate"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0.0
    assert thermostat.get_requested_compressor_speed() == 0.0
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [88888888]


async def test_xl824_2(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get methods for an xl824 thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_xl624.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2222222, 3333333]
    thermostat = nexia.get_thermostat_by_id(3333333)

    assert thermostat.get_model() == "XL824"
    assert thermostat.get_firmware() == "5.9.1"
    assert thermostat.get_dev_build_number() == "1581314625"
    assert thermostat.get_device_id() == "01573380"
    assert thermostat.get_type() == "XL824"
    assert thermostat.get_name() == "Upstairs"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Circulate"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0.0
    assert thermostat.get_requested_compressor_speed() == 0.0
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [99999999]


async def test_basic(aiohttp_session):
    """Basic tests for NexiaHome."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)

    assert nexia.get_name() == "Hidden"
    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2059661, 2059676, 2293892, 2059652]


async def test_basic_issue_33758(aiohttp_session):
    """Basic tests for NexiaHome."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33758.json"))
    nexia.update_from_json(devices_json)

    assert nexia.get_name() == "Hidden"
    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [12345678]


async def test_zone_issue_33968_zone_83037337(aiohttp_session):
    """Tests for nexia thermostat zone that is cooling."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33968.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(1690380)
    zone = thermostat.get_zone_by_id(83037337)

    assert zone.thermostat == thermostat

    assert zone.get_name() == "Family Room"
    assert zone.get_cooling_setpoint() == 77
    assert zone.get_heating_setpoint() == 74
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Damper Closed"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is False
    assert zone.is_in_permanent_hold() is True


async def test_zone_issue_33968_zone_83037340(aiohttp_session):
    """Tests for nexia thermostat zone that is cooling."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33968.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(1690380)
    zone = thermostat.get_zone_by_id(83037340)

    assert zone.thermostat == thermostat

    assert zone.get_name() == "Office"
    assert zone.get_cooling_setpoint() == 77
    assert zone.get_heating_setpoint() == 74
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Damper Open"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_zone_issue_33968_zone_83037343(
    aiohttp_session: aiohttp.ClientSession,
) -> None:
    """Tests for nexia thermostat zone that is cooling."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33968.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(1690380)
    zone = thermostat.get_zone_by_id(83037343)

    assert zone.thermostat == thermostat

    assert zone.get_name() == "Master"
    assert zone.get_cooling_setpoint() == 77
    assert zone.get_heating_setpoint() == 68
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Damper Open"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_zone_issue_33758(aiohttp_session: aiohttp.ClientSession) -> None:
    """Tests for nexia thermostat zone relieving air."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33758.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(12345678)
    zone = thermostat.get_zone_by_id(12345678)

    assert zone.thermostat == thermostat

    assert zone.get_name() == "Thermostat NativeZone"
    assert zone.get_cooling_setpoint() == 73
    assert zone.get_heating_setpoint() == 68
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Run Schedule - None"
    assert zone.is_calling() is False
    assert zone.is_in_permanent_hold() is False


async def test_zone_relieving_air(aiohttp_session):
    """Tests for nexia thermostat zone relieving air."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(2293892)
    zone = thermostat.get_zone_by_id(83394133)

    assert zone.thermostat == thermostat

    assert zone.get_name() == "Bath Closet"
    assert zone.get_cooling_setpoint() == 79
    assert zone.get_heating_setpoint() == 63
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Relieving Air"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_zone_cooling_air(aiohttp_session):
    """Tests for nexia thermostat zone cooling."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(2293892)
    zone = thermostat.get_zone_by_id(83394130)

    assert zone.get_name() == "Master"
    assert zone.get_cooling_setpoint() == 71
    assert zone.get_heating_setpoint() == 63
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Damper Open"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_zone_idle(aiohttp_session):
    """Tests for nexia thermostat zone idle."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_houses_123456.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(2059661)
    zone = thermostat.get_zone_by_id(83261002)

    assert zone.get_name() == "Living East"
    assert zone.get_cooling_setpoint() == 79
    assert zone.get_heating_setpoint() == 63
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is False
    assert zone.is_in_permanent_hold() is True


async def test_xl824_idle(aiohttp_session):
    """Tests for nexia xl824 zone idle."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_xl624.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2222222, 3333333]
    thermostat = nexia.get_thermostat_by_id(3333333)
    zone = thermostat.get_zone_by_id(99999999)

    assert zone.get_name() == "Upstairs NativeZone"
    assert zone.get_cooling_setpoint() == 74
    assert zone.get_heating_setpoint() == 62
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is False
    assert zone.is_in_permanent_hold() is True


async def test_single_zone(aiohttp_session: aiohttp.ClientSession) -> None:
    """Test thermostat with only a single (Native) zone."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("single_zone_xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [345678]
    thermostat = nexia.get_thermostat_by_id(345678)
    zone = thermostat.get_zone_by_id(234567)

    assert zone.get_name() == "Thermostat 1 NativeZone"
    assert zone.get_cooling_setpoint() == 73
    assert zone.get_heating_setpoint() == 68
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True

    # No sensors
    sensors = zone.get_sensors()
    assert len(sensors) == 0


async def test_single_zone_system_off(aiohttp_session: aiohttp.ClientSession) -> None:
    """Test thermostat with only a single (Native) zone."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("single_zone_xl1050_system_off.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [345678]
    thermostat = nexia.get_thermostat_by_id(345678)
    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.3"
    assert thermostat.get_dev_build_number() == "1599485560"
    assert thermostat.get_device_id() == "028E05EC"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Thermostat 1"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.get_variable_fan_speed_limits() == (0.35, 1.0)
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode(), "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_outdoor_temperature() == 64.0
    assert thermostat.get_relative_humidity() == 0.59
    assert thermostat.get_current_compressor_speed() == 0.0
    assert thermostat.get_requested_compressor_speed() == 0.0
    assert thermostat.get_fan_speed_setpoint() == 0.5
    assert thermostat.get_dehumidify_setpoint() == 0.50
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is True
    assert thermostat.get_system_status() == "System Off"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.get_air_cleaner_mode() == "auto"
    assert thermostat.is_blower_active() is False

    zone = thermostat.get_zone_by_id(234567)

    assert zone.get_name() == "Thermostat 1 NativeZone"
    assert zone.get_cooling_setpoint() == 73
    assert zone.get_heating_setpoint() == 68
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_automations(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Get methods for an active thermostat."""
    nexia = NexiaHome(aiohttp_session)
    text = await load_fixture("mobile_houses_123456.json")
    devices_json = json.loads(text)
    nexia.update_from_json(devices_json)

    automation_ids = nexia.get_automation_ids()
    assert automation_ids == [
        3467876,
        3467870,
        3452469,
        3452472,
        3454776,
        3454774,
        3486078,
        3486091,
    ]

    automation_one = nexia.get_automation_by_id(3467876)

    assert automation_one.name == "Away for 12 Hours"
    assert automation_one.description == (
        "When IFTTT activates the automation Upstairs West Wing will "
        "permanently hold the heat to 62.0 and cool to 83.0 AND "
        "Downstairs East Wing will permanently hold the heat to 62.0 "
        "and cool to 83.0 AND Downstairs West Wing will permanently "
        "hold the heat to 62.0 and cool to 83.0 AND Activate the mode "
        "named 'Away 12' AND Master Suite will permanently hold the "
        "heat to 62.0 and cool to 83.0"
    )
    assert automation_one.enabled is True
    assert automation_one.automation_id == 3467876

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/automations/3467876/activate",
        payload={
            "success": True,
            "error": None,
        },
    )
    await automation_one.activate()


async def test_x850_grouped(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get methods for an xl850 grouped thermostat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("grouped_xl850.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2323232]
    thermostat = nexia.get_thermostat_by_id(2323232)

    assert thermostat.get_model() == "XL850"
    assert thermostat.get_firmware() == "5.9.7"
    assert thermostat.get_dev_build_number() == "XXXXXXXXXXXXXXXXX"
    assert thermostat.get_device_id() == "XXXXXXXXXXX"
    assert thermostat.get_type() == "XL850"
    assert thermostat.get_name() == "Hallway Control"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.10, 0.65)
    assert thermostat.get_fan_mode() == "Circulate"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 1.0
    assert thermostat.get_requested_compressor_speed() == 1.0
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is True
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "Cooling"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is True

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [343434334]
    zone = thermostat.get_zone_by_id(343434334)

    assert zone.get_name() == "Hallway Control NativeZone"
    assert zone.get_cooling_setpoint() == 75
    assert zone.get_heating_setpoint() == 68
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_issue_79891(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get methods issue 79891 thermostat.

    https://github.com/home-assistant/core/issues/79891
    """
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("issue_79891.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2473073, 2222249]

    # Thermostat 1
    thermostat = nexia.get_thermostat_by_id(2473073)
    assert thermostat.get_model() == "XL824"
    assert thermostat.get_firmware() == "5.9.6"
    assert thermostat.get_dev_build_number() == "1614575015"
    assert thermostat.get_device_id() == "01695948"
    assert thermostat.get_type() == "XL824"
    assert thermostat.get_name() == "First floor"
    assert thermostat.get_deadband() == 2
    assert thermostat.get_setpoint_limits() == (13.0, 37.0)
    assert thermostat.has_variable_fan_speed() is False
    assert thermostat.get_unit() == "C"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0
    assert thermostat.get_requested_compressor_speed() == 0
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83496154]
    zone = thermostat.get_zone_by_id(83496154)

    assert zone.get_name() == "First floor NativeZone"
    assert zone.get_cooling_setpoint() == 26.5
    assert zone.get_heating_setpoint() == 22.0
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Run Schedule - None"
    assert zone.is_calling() is False
    assert zone.is_in_permanent_hold() is False

    # Thermostat 2
    thermostat = nexia.get_thermostat_by_id(2222249)
    assert thermostat.get_model() == "XR724"
    assert thermostat.get_firmware() == "02.72.00"
    assert thermostat.get_dev_build_number() is None
    assert thermostat.get_device_id() == "0080E1CB6FF0"
    assert thermostat.get_type() == "XR724"
    assert thermostat.get_name() == "Second floor"
    assert thermostat.get_deadband() == 2
    assert thermostat.get_setpoint_limits() == (13.0, 37.0)
    assert thermostat.has_variable_fan_speed() is False
    assert thermostat.get_unit() == "C"
    assert thermostat.get_humidity_setpoint_limits() == (0.10, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Circulate", "Auto", "On"]
    assert thermostat.get_current_compressor_speed() == 0
    assert thermostat.get_requested_compressor_speed() == 0
    assert thermostat.has_dehumidify_support() is False
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is False
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [1]
    zone = thermostat.get_zone_by_id(1)

    assert zone.get_name() == "default"
    assert zone.get_cooling_setpoint() == 27.0
    assert zone.get_heating_setpoint() == 20.0
    assert zone.get_current_mode() == "AUTO"
    assert zone.get_requested_mode() == "AUTO"
    assert zone.get_presets() == ["Off", "Auto", "Cool", "Heat"]
    assert zone.get_preset() == "Auto"
    assert zone.get_status() == "auto"
    assert zone.get_setpoint_status() == "Run Schedule - Auto"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is False


async def test_new_xl1050(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get with a new xl1050."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [4122267, 2059676, 4125886, 2059652]

    # Thermostat 1
    thermostat = nexia.get_thermostat_by_id(4122267)
    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.6"
    assert thermostat.get_dev_build_number() == "1614588140"
    assert thermostat.get_device_id() == "02868F20"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Downstairs East Wing"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0
    assert thermostat.get_requested_compressor_speed() == 0
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "System Idle"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [84398305, 84398308, 84398311, 84398314]
    zone = thermostat.get_zone_by_id(84398305)

    assert zone.get_name() == "Living East"
    assert zone.get_cooling_setpoint() == 75
    assert zone.get_heating_setpoint() == 55
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_temperature() == 72
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Run Schedule - None"
    assert zone.is_calling() is False
    zone.check_heat_cool_setpoints(70, 76)
    assert zone.is_in_permanent_hold() is False

    # Sensors
    thermostat = nexia.get_thermostat_by_id(2059676)

    assert thermostat.get_zone_ids() == [83261015, 83261018, 84243806, 83535112]
    zone = thermostat.get_zone_by_id(83261015)
    sensors = zone.get_sensors()

    assert len(sensors) == 1
    sensor = sensors[0]
    assert sensor.id == 16800389
    assert sensor.name == "Living West"
    assert sensor.type == "thermostat"
    assert sensor.serial_number == "NativeIDTUniqueID"
    assert sensor.weight == 1.0
    assert sensor.temperature == 70
    assert sensor.temperature_valid is True
    assert sensor.humidity == 42
    assert sensor.humidity_valid is True
    assert sensor.has_online is False
    assert sensor.has_battery is False


async def test_new_xl824(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get with a new xl824."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl824.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [1841910]

    # Thermostat 1
    thermostat = nexia.get_thermostat_by_id(1841910)
    assert thermostat.get_model() == "XL824"
    assert thermostat.get_firmware() == "5.9.6"
    assert thermostat.get_dev_build_number() == "1614575015"
    assert thermostat.get_device_id() == "0160F840"
    assert thermostat.get_type() == "XL824"
    assert thermostat.get_name() == "Office Thermostat"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.1, 0.45)
    assert thermostat.get_fan_mode() == "On"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0
    assert thermostat.get_requested_compressor_speed() == 0
    assert thermostat.has_dehumidify_support() is False
    assert thermostat.has_humidify_support() is True
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "Fan Running"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is True
    assert thermostat.is_online is True

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83128724]
    zone = thermostat.get_zone_by_id(83128724)

    assert zone.get_name() == "Office Thermostat NativeZone"
    assert zone.get_cooling_setpoint() == 99
    assert zone.get_heating_setpoint() == 55
    assert zone.get_current_mode() == "OFF"
    assert zone.get_requested_mode() == "OFF"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_system_offline(aiohttp_session: aiohttp.ClientSession) -> None:
    """Get a system offline."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("system_offline.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [2224092, 3887969, 3295333]

    # Thermostat 1
    thermostat = nexia.get_thermostat_by_id(2224092)
    assert thermostat.get_model() == "XL1050"
    assert thermostat.get_firmware() == "5.9.6"
    assert thermostat.get_dev_build_number() == "1614588140"
    assert thermostat.get_device_id() == "02863D94"
    assert thermostat.get_type() == "XL1050"
    assert thermostat.get_name() == "Game Room"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.35, 0.65)
    assert thermostat.get_fan_mode() == "Auto"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0.35
    assert thermostat.get_requested_compressor_speed() == 0.35
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is False
    assert thermostat.has_emergency_heat() is False
    assert thermostat.get_system_status() == "NOT CONNECTED"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is True
    assert thermostat.is_online is False

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [83354736, 83354739, 83354742, 83354745]
    zone = thermostat.get_zone_by_id(83354736)

    assert zone.get_name() == "Game Room"
    assert zone.get_cooling_setpoint() == 75
    assert zone.get_heating_setpoint() == 55
    assert zone.get_current_mode() == "COOL"
    assert zone.get_requested_mode() == "COOL"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Damper Open"
    assert zone.get_setpoint_status() == "Permanent Hold"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is True


async def test_emergency_heat(aiohttp_session: aiohttp.ClientSession) -> None:
    """Test emergency heat."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("eme_heat.json"))
    nexia.update_from_json(devices_json)

    thermostat_ids = nexia.get_thermostat_ids()
    assert thermostat_ids == [3983351]

    # Thermostat 1
    thermostat = nexia.get_thermostat_by_id(3983351)
    assert thermostat.get_model() == "XL850"
    assert thermostat.get_firmware() == "5.9.6"
    assert thermostat.get_dev_build_number() == "1614581867"
    assert thermostat.get_device_id() == "00D68470"
    assert thermostat.get_type() == "XL850"
    assert thermostat.get_name() == "XL850 Home"
    assert thermostat.get_deadband() == 3
    assert thermostat.get_setpoint_limits() == (55, 99)
    assert thermostat.has_variable_fan_speed() is True
    assert thermostat.get_unit() == "F"
    assert thermostat.get_humidity_setpoint_limits() == (0.10, 0.65)
    assert thermostat.get_fan_mode() == "On"
    assert thermostat.get_fan_modes() == ["Auto", "On", "Circulate"]
    assert thermostat.get_current_compressor_speed() == 0
    assert thermostat.get_requested_compressor_speed() == 0
    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is True
    assert thermostat.has_emergency_heat() is True
    assert thermostat.get_system_status() == "Fan Running"
    assert thermostat.has_air_cleaner() is True
    assert thermostat.is_blower_active() is True
    assert thermostat.is_online is True

    zone_ids = thermostat.get_zone_ids()
    assert zone_ids == [84326108]
    zone = thermostat.get_zone_by_id(84326108)

    assert zone.get_name() == "XL850 Home NativeZone"
    assert zone.get_cooling_setpoint() == 99
    assert zone.get_heating_setpoint() == 68
    assert zone.get_current_mode() == "HEAT"
    assert zone.get_requested_mode() == "HEAT"
    assert zone.get_presets() == ["None", "Home", "Away", "Sleep"]
    assert zone.get_preset() == "None"
    assert zone.get_status() == "Idle"
    assert zone.get_setpoint_status() == "Run Schedule - None"
    assert zone.is_calling() is True
    assert zone.is_in_permanent_hold() is False


async def test_humidity_and_fan_mode(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Tests for preventing an API timeout when updating humidity
    and fan modes to the same value
    """
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_issue_33758.json"))
    nexia.update_from_json(devices_json)

    thermostat: NexiaThermostat = nexia.get_thermostat_by_id(12345678)
    devices = _extract_devices_from_houses_json(devices_json)

    assert thermostat.has_dehumidify_support() is True
    assert thermostat.has_humidify_support() is True
    assert thermostat.get_humidify_setpoint_limits() == (0.10, 0.45)
    assert thermostat.get_dehumidify_setpoint_limits() == (0.35, 0.65)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/fan_mode",
        payload={"result": devices[0]},
    )
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/humidify",
        payload={"result": devices[0]},
    )
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/dehumidify",
        payload={"result": devices[0]},
    )

    with pytest.raises(KeyError, match="Invalid fan mode.*"):
        await thermostat.set_fan_mode("DOES_NOT_EXIST")

    # Attempting to set to the same value should not trigger an API call
    await thermostat.set_fan_mode("Auto")
    assert not mock_aioresponse.requests.get(
        ("POST", "https://www.mynexia.com/mobile/xxl_thermostats/12345678/fan_mode")
    )

    # Attempting to set to different value should trigger an API call
    await thermostat.set_fan_mode("On")
    assert mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/fan_mode"),
        )
    )

    # Attempting to set to the same value should not trigger an API call
    await thermostat.set_humidity_setpoints(
        humidify_setpoint=0.4, dehumidify_setpoint=0.55
    )
    assert not mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/humidify"),
        )
    )
    assert not mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/dehumidify"),
        )
    )

    # Attempting to set to different value should trigger an API call
    await thermostat.set_humidity_setpoints(
        humidify_setpoint=0.15, dehumidify_setpoint=0.60
    )
    request = mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/humidify"),
        )
    )
    assert request is not None
    first_request = request[0]
    assert first_request.kwargs["json"]["value"] == "0.15"
    request = mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/dehumidify"),
        )
    )
    assert request is not None
    first_request = request[0]
    assert first_request.kwargs["json"]["value"] == "0.6"

    mock_aioresponse.requests.clear()
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/humidify",
        payload={"result": devices[0]},
    )
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_thermostats/12345678/dehumidify",
        payload={"result": devices[0]},
    )
    # Attempting to set to an out of range value should clamp to the nearest valid value
    await thermostat.set_humidity_setpoints(
        humidify_setpoint=0.242, dehumidify_setpoint=0.652
    )
    request = mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/humidify"),
        )
    )
    assert request is not None
    first_request = request[0]
    assert first_request.kwargs["json"]["value"] == "0.25"
    request = mock_aioresponse.requests.get(
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_thermostats/12345678/dehumidify"),
        )
    )
    assert request is not None
    first_request = request[0]
    assert first_request.kwargs["json"]["value"] == "0.65"


async def test_sensor_access(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test sensor access methods."""
    persist_file = Path("nexia_config_test.conf")
    nexia = NexiaHome(aiohttp_session, house_id=2582941, state_file=persist_file)
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/accounts/sign_in",
        payload={
            "success": True,
            "error": None,
            "result": {
                "mobile_id": 5400000,
                "api_key": "10654c0be00000000000000000000000",
                "setup_step": "done",
                "locale": "en_us",
            },
        },
    )
    await nexia.login()

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/houses/2582941",
        body=await load_fixture("sensors_xl1050_house.json"),
    )
    assert await nexia.update() is not None

    assert nexia.get_thermostat_ids() == [5378307]
    thermostat: NexiaThermostat = nexia.get_thermostat_by_id(5378307)

    assert thermostat.get_zone_ids() == [85034552]
    zone = thermostat.get_zone_by_id(85034552)
    sensors = zone.get_sensors()

    assert len(sensors) == 2
    sensor = sensors[0]
    assert sensor.id == 17687546
    assert sensor.name == "Center"
    assert sensor.type == "thermostat"
    assert sensor.serial_number == "NativeIDTUniqueID"
    assert sensor.weight == 0.5
    assert sensor.temperature == 68
    assert sensor.temperature_valid is True
    assert sensor.humidity == 32
    assert sensor.humidity_valid is True
    assert sensor.has_online is False
    assert sensor.connected is None
    assert sensor.has_battery is False
    assert sensor.battery_level is None
    assert sensor.battery_low is None
    assert sensor.battery_valid is None

    sensor = sensors[1]
    assert sensor.id == 17687549
    assert sensor.name == "Upstairs"
    assert sensor.type == "930"
    assert sensor.serial_number == "2410R5C53X"
    assert sensor.weight == 0.5
    assert sensor.temperature == 69
    assert sensor.temperature_valid is True
    assert sensor.humidity == 32
    assert sensor.humidity_valid is True
    assert sensor.has_online is True
    assert sensor.connected is True
    assert sensor.has_battery is True
    assert sensor.battery_level == 95
    assert sensor.battery_low is False
    assert sensor.battery_valid is True

    # execute log response code path
    nexia.log_response = True

    # execute no completion code path
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/85034552/request_current_sensor_state",
        payload={
            "success": True,
            "error": None,
            "result": {
                "polling_path": "https://www.mynexia.com/backstage/announcements/6a31e745716789b84603036489fe8d1e35ca80fa50000000"
            },
        },
    )
    assert await zone.load_current_sensor_state(max_polls=0) is False

    # execute normal code path
    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/85034552/request_current_sensor_state",
        payload={
            "success": True,
            "error": None,
            "result": {
                "polling_path": "https://www.mynexia.com/backstage/announcements/6a31e745716789b84603036489fe8d1e35ca80fa5dd381e5"
            },
        },
    )
    mock_aioresponse.get(
        "https://www.mynexia.com/backstage/announcements/6a31e745716789b84603036489fe8d1e35ca80fa5dd381e5",
        body=b"null",
    )
    mock_aioresponse.get(
        "https://www.mynexia.com/backstage/announcements/6a31e745716789b84603036489fe8d1e35ca80fa5dd381e5",
        payload={
            "status": "success, altered to enhance test coverage",
            "options": {},
        },
    )
    assert await zone.load_current_sensor_state(0.01) is True

    mock_aioresponse.get(
        "https://www.mynexia.com/mobile/xxl_thermostats/5378307",
        body=await load_fixture("sensors_xl1050_thermostat.json"),
    )
    await thermostat.refresh_thermostat_data()
    sensors = zone.get_sensors()

    assert len(sensors) == 2
    sensor = sensors[0]
    assert sensor.id == 17687546
    assert sensor.name == "Center"
    assert sensor.weight == 0.5
    assert sensor.temperature == 69
    assert sensor.humidity == 33

    sensor = sensors[1]
    assert sensor.id == 17687549
    assert sensor.name == "Upstairs"
    assert sensor.weight == 0.5
    assert sensor.temperature == 70
    assert sensor.humidity == 33

    assert persist_file.exists() is True
    persist_file.unlink()
    assert persist_file.exists() is False


async def test_clamp_to_predefined_values() -> None:
    assert clamp_to_predefined_values(45, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 50
    assert clamp_to_predefined_values(50, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 50
    assert clamp_to_predefined_values(51, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 50
    assert clamp_to_predefined_values(52, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 50
    assert clamp_to_predefined_values(53, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 55
    assert clamp_to_predefined_values(52.51, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 55
    assert clamp_to_predefined_values(55, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 55
    assert clamp_to_predefined_values(56, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 55
    assert clamp_to_predefined_values(90, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 90
    assert clamp_to_predefined_values(95, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 90
    assert clamp_to_predefined_values(100, [50, 55, 60, 65, 70, 75, 80, 85, 90]) == 90
    assert clamp_to_predefined_values(100, [90, 85, 80, 75, 70]) == 90
    assert clamp_to_predefined_values(0.4, [0.1, 0.2, 0.3, 0.4, 0.5]) == 0.4
    assert clamp_to_predefined_values(0.45, [0.1, 0.2, 0.3, 0.4, 0.5]) == 0.4


async def test_set_preset(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test setting a zone preset."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(4122267)
    zone = thermostat.get_zone_by_id(84398305)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 84398305)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/84398305/preset_selected",
        payload={"result": zone_data},
    )
    await zone.set_preset("Home")

    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/84398305/preset_selected"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"]["value"] == 1

    mock_aioresponse.requests.clear()

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/84398305/preset_selected",
        payload={"result": zone_data},
    )
    await zone.set_preset("Away")

    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/84398305/preset_selected"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"]["value"] == 2

    mock_aioresponse.requests.clear()


async def test_set_return_to_schedule_already_in_schedule(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test returning to schedule."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(4122267)
    zone = thermostat.get_zone_by_id(84398305)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 84398305)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/84398305/return_to_schedule",
        payload={"result": zone_data},
    )
    await zone.call_return_to_schedule()

    assert not mock_aioresponse.requests  # already in schedule


async def test_set_return_to_schedule_from_hold(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test returning to schedule."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(4122267)
    zone = thermostat.get_zone_by_id(84398314)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 84398314)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/84398314/run_mode",
        payload={"result": zone_data},
    )
    await zone.call_return_to_schedule()
    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/84398314/run_mode"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {"value": "run_schedule"}


async def test_set_return_to_schedule_xl824(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test returning to schedule with xl824."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl824.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(1841910)
    zone = thermostat.get_zone_by_id(83128724)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 83128724)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/83128724/run_mode",
        payload={"result": zone_data},
    )
    await zone.call_return_to_schedule()
    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/83128724/run_mode"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {"value": "run_schedule"}


async def test_set_return_to_schedule_single_zone_xl1050(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test returning to schedule with xl1050 single zone."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("single_zone_xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(345678)
    zone = thermostat.get_zone_by_id(234567)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 234567)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/234567/return_to_schedule",
        payload={"result": zone_data},
    )
    await zone.call_return_to_schedule()
    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/234567/return_to_schedule"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {}


async def test_set_return_to_schedule_single_zone_xl624(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test returning to schedule with xl624."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("mobile_house_xl624.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(2222222)
    zone = thermostat.get_zone_by_id(88888888)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[1]["zones"], "id", 88888888)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/88888888/return_to_schedule",
        payload={"result": zone_data},
    )
    await zone.call_return_to_schedule()
    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/88888888/return_to_schedule"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {}


async def test_set_permanent_hold(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test returning to schedule."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(4122267)
    zone = thermostat.get_zone_by_id(84398305)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 84398305)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/84398305/run_mode",
        payload={"result": zone_data},
    )
    await zone.call_permanent_hold()

    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/84398305/run_mode"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {
        "value": "permanent_hold",
    }


async def test_set_zone_mode(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test setting zone mode."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("xl1050.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id(4122267)
    zone = thermostat.get_zone_by_id(84398305)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 84398305)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/xxl_zones/84398305/zone_mode",
        payload={"result": zone_data},
    )
    await zone.set_mode("AUTO")

    requests = mock_aioresponse.requests[
        (
            "POST",
            URL("https://www.mynexia.com/mobile/xxl_zones/84398305/zone_mode"),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {
        "value": "AUTO",
    }


async def test_set_perm_hold_ux360(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test perm hold for a ux360."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("ux360.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id("123456")
    zone = thermostat.get_zone_by_id(1)

    devices = _extract_devices_from_houses_json(devices_json)
    children = extract_children_from_devices_json(devices)
    zone_data = find_dict_with_keyvalue_in_json(children[0]["zones"], "id", 1)

    assert zone.is_in_permanent_hold() is False

    await zone.call_return_to_schedule()

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/diagnostics/thermostats/123456/run_mode/1",
        payload={"result": zone_data},
    )
    await zone.call_permanent_hold()

    requests = mock_aioresponse.requests[
        (
            "POST",
            URL(
                "https://www.mynexia.com/mobile/diagnostics/thermostats/123456/run_mode/1"
            ),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {"value": "hold"}


async def test_set_fan_mode_ux360(
    aiohttp_session: aiohttp.ClientSession, mock_aioresponse: aioresponses
) -> None:
    """Test fan mode on a ux360."""
    nexia = NexiaHome(aiohttp_session)
    devices_json = json.loads(await load_fixture("ux360.json"))
    nexia.update_from_json(devices_json)

    thermostat = nexia.get_thermostat_by_id("123456")

    devices = _extract_devices_from_houses_json(devices_json)

    mock_aioresponse.post(
        "https://www.mynexia.com/mobile/diagnostics/thermostats/123456/fan_mode",
        payload={"result": devices[0]},
    )
    await thermostat.set_fan_mode("Circulate")

    requests = mock_aioresponse.requests[
        (
            "POST",
            URL(
                "https://www.mynexia.com/mobile/diagnostics/thermostats/123456/fan_mode"
            ),
        )
    ]
    assert requests is not None
    first_request = requests[0]
    assert first_request.kwargs["json"] == {"value": "circulate"}



"""Nexia Thermostat."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Any

from .const import (
    AIR_CLEANER_MODES,
    BLOWER_OFF_STATUSES,
    DEFAULT_UPDATE_METHOD,
    HUMIDITY_MAX,
    HUMIDITY_MIN,
)
from .util import find_dict_with_keyvalue_in_json, find_humidity_setpoint, is_number
from .zone import NexiaThermostatZone

_LOGGER = logging.getLogger(__name__)


if TYPE_CHECKING:
    from .home import NexiaHome


def clamp_to_predefined_values(target: float, values: list[float]) -> float:
    """Clamp a target value to the nearest predefined value."""
    if target in values:
        return target
    return values[min(range(len(values)), key=lambda i: abs(values[i] - target))]


class ThermostatEndpoint(Enum):
    """Enum for Thermostat Endpoints."""

    FAN_MODE = auto()
    FAN_SPEED = auto()
    AIR_CLEANER_MODE = auto()
    SCHEDULING_ENABLED = auto()
    EMERGENCY_HEAT = auto()
    DEHUMIDIFY = auto()
    HUMIDIFY = auto()


@dataclass
class ThermostatEndPointData:
    """Dataclass for Thermostat Endpoints.

    area/area_primary_key/key are used to find the endpoint in the thermostat json.
    action is the action to take on the endpoint.
    fallback_endpoint is the endpoint to use if the action is not found.
    """

    area: str
    area_primary_key: str
    key: str
    action: str  # always looks for `self` first
    fallback_endpoint: str


ENDPOINT_MAP = {
    ThermostatEndpoint.FAN_MODE: ThermostatEndPointData(
        area="features",
        area_primary_key="name",
        key="thermostat_fan_mode",
        action="update_thermostat_fan_mode",
        fallback_endpoint="fan_mode",
    ),
    ThermostatEndpoint.FAN_SPEED: ThermostatEndPointData(
        area="settings",
        area_primary_key="type",
        key="fan_speed",
        action="fan_speed",
        fallback_endpoint="fan_speed",
    ),
    ThermostatEndpoint.AIR_CLEANER_MODE: ThermostatEndPointData(
        area="settings",
        area_primary_key="type",
        key="air_cleaner_mode",
        action="air_cleaner_mode",
        fallback_endpoint="air_cleaner_mode",
    ),
    ThermostatEndpoint.SCHEDULING_ENABLED: ThermostatEndPointData(
        area="settings",
        area_primary_key="type",
        key="scheduling_enabled",
        action="scheduling_enabled",
        fallback_endpoint="scheduling_enabled",
    ),
    ThermostatEndpoint.EMERGENCY_HEAT: ThermostatEndPointData(
        area="settings",
        area_primary_key="type",
        key="emergency_heat",
        action="emergency_heat",
        fallback_endpoint="emergency_heat",
    ),
    ThermostatEndpoint.DEHUMIDIFY: ThermostatEndPointData(
        area="settings",
        area_primary_key="type",
        key="dehumidify",
        action="dehumidify",
        fallback_endpoint="dehumidify",
    ),
    ThermostatEndpoint.HUMIDIFY: ThermostatEndPointData(
        area="settings",
        area_primary_key="type",
        key="humidify",
        action="humidify",
        fallback_endpoint="humidify",
    ),
}


class NexiaThermostat:
    """A nexia Thermostat.

    Represents a nexia thermostat.
    """

    def __init__(self, nexia_home: NexiaHome, thermostat_json: dict[str, Any]) -> None:
        """Init nexia Thermostat."""
        self._nexia_home = nexia_home
        self.thermostat_id: int | str = thermostat_json["id"]
        self._thermostat_json = thermostat_json
        if self.has_zones():
            self.zones = [
                NexiaThermostatZone(nexia_home, self, zone)
                for zone in thermostat_json["zones"]
            ]
        else:
            self.zones = []

    @property
    def API_MOBILE_THERMOSTAT_URL(self) -> str:  # pylint: disable=invalid-name
        return (
            self._nexia_home.mobile_url + "/xxl_thermostats/{thermostat_id}/{end_point}"
        )

    @property
    def is_online(self) -> bool:
        """Returns whether the thermostat is online or not.
        :return: bool.
        """
        return self.get_system_status().upper() != "NOT CONNECTED"

    def _get_thermostat_advanced_info_label(self, label: str) -> str | None:
        """Lookup advanced_info in the thermostat features and find the value of the
        requested label.
        """
        advanced_info = self._get_thermostat_features_key("advanced_info")

        try:
            return find_dict_with_keyvalue_in_json(
                advanced_info["items"],
                "label",
                label,
            )["value"]
        except KeyError:
            return None

    def get_model(self) -> str | None:
        """Returns the thermostat model
        :return: string.
        """
        return self._get_thermostat_advanced_info_label("Model")

    def get_firmware(self) -> str | None:
        """Returns the thermostat firmware version
        :return: string.
        """
        return self._get_thermostat_advanced_info_label(
            "Firmware Version",
        ) or self._get_thermostat_advanced_info_label("Main Firmware Version")

    def get_dev_build_number(self) -> str | None:
        """Returns the thermostat development build number.
        :return: string.
        """
        return self._get_thermostat_advanced_info_label(
            "Firmware Build Number",
        ) or self._get_thermostat_advanced_info_label("Version")

    def get_device_id(self) -> str | None:
        """Returns the device id
        :return: string.
        """
        return self._get_thermostat_advanced_info_label("AUID")

    def get_type(self) -> str | None:
        """Returns the thermostat type, such as TraneXl1050
        :return: str.
        """
        return self.get_model()

    def get_name(self) -> str:
        """Returns the name of the thermostat. This is not the zone name.
        :return: str.
        """
        return self._get_thermostat_key("name")

    ########################################################################
    # Supported Features

    def has_outdoor_temperature(self) -> bool:
        """Capability indication of whether the thermostat has an outdoor
        temperature sensor
        :return: bool.
        """
        return self._get_thermostat_key_or_none("has_outdoor_temperature")

    def has_relative_humidity(self) -> bool:
        """Capability indication of whether the thermostat has a relative
        humidity sensor
        :return: bool.
        """
        return bool(self._get_thermostat_key_or_none("indoor_humidity"))

    def has_variable_speed_compressor(self) -> bool:
        """Capability indication of whether the thermostat has a variable speed
        compressor
        :return: bool.
        """
        # This only shows up if it's running on mobile
        return True

    def has_emergency_heat(self) -> bool:
        """Capability indication of whether the thermostat has emergency/aux heat.
        :return: bool.
        """
        return bool(self.get_thermostat_settings_key_or_none("emergency_heat"))

    def has_variable_fan_speed(self) -> bool:
        """Capability indication of whether the thermostat has a variable speed
        blower
        :return: bool.
        """
        return bool(self.get_thermostat_settings_key_or_none("fan_speed"))

    def has_zones(self) -> bool:
        """Indication of whether zoning is enabled or not on the thermostat.
        :return: bool.
        """
        return bool(self._get_thermostat_key_or_none("zones"))

    def has_dehumidify_support(self) -> bool:
        """Indication of whether dehumidifying support is available.
        :return: bool.
        """
        return bool(self.get_thermostat_settings_key_or_none("dehumidify"))

    def has_humidify_support(self) -> bool:
        """Indication of whether humidifying support is available.
        :return: bool.
        """
        return bool(self.get_thermostat_settings_key_or_none("humidify"))

    ########################################################################
    # System Attributes

    def get_deadband(self) -> int:
        """Returns the deadband of the thermostat. This is the minimum number of
        degrees between the heat and cool setpoints in the number of degrees in
        the temperature unit selected by the
        thermostat.
        :return: int.
        """
        return self._get_thermostat_features_key("thermostat")["setpoint_delta"]

    def get_setpoint_limits(self) -> tuple[int, int]:
        """Returns a tuple of the minimum and maximum temperature that can be set
        on any zone. This is in the temperature unit selected by the
        thermostat.
        :return: (int, int).
        """
        return (
            self._get_thermostat_features_key("thermostat")["setpoint_heat_min"],
            self._get_thermostat_features_key("thermostat")["setpoint_cool_max"],
        )

    def get_variable_fan_speed_limits(self) -> tuple[float, float]:
        """Returns the variable fan speed setpoint limits of the thermostat.
        :return: (float, float).
        """
        if self.has_variable_fan_speed():
            possible_values = self.get_thermostat_settings_key("fan_speed")["values"]
            return (possible_values[0], possible_values[-1])
        raise AttributeError("This thermostat does not support fan speeds")

    def get_unit(self) -> str:
        """Returns the temperature unit used by this system, either C or F.
        :return: str.
        """
        return self._get_thermostat_features_key("thermostat")["scale"].upper()

    @property
    def humidify_setpoints(self) -> list[float]:
        """Returns the humidify setpoints of the thermostat.
        :return: list[float].
        """
        return self._get_thermostat_deep_key("settings", "type", "humidify")["values"]

    @property
    def dehumidify_setpoints(self) -> list[float]:
        """Returns the dehumidify setpoints of the thermostat.
        :return: list[float].
        """
        return self._get_thermostat_deep_key("settings", "type", "dehumidify")["values"]

    def get_humidify_setpoint_limits(self) -> tuple[float, float]:
        """Returns humidify setpoint limits of the thermostat.
        :return: (float, float)
        """
        humidify_values = self.humidify_setpoints
        return min(humidify_values), max(humidify_values)

    def get_dehumidify_setpoint_limits(self) -> tuple[float, float]:
        """Returns dehumidify setpoint limits of the thermostat.
        :return: (float, float)
        """
        dehumidify_values = self.dehumidify_setpoints
        return min(dehumidify_values), max(dehumidify_values)

    def get_humidity_setpoint_limits(self) -> tuple[float, float]:
        """Returns the humidity setpoint limits of the thermostat
        :return: (float, float)
        """
        if self.has_humidify_support() and self.has_dehumidify_support():
            return min(self.get_humidify_setpoint_limits()), max(
                self.get_dehumidify_setpoint_limits()
            )
        if self.has_humidify_support():
            return self.get_humidify_setpoint_limits()
        if self.has_dehumidify_support():
            return self.get_dehumidify_setpoint_limits()
        # Fall back to hard coded limits
        return HUMIDITY_MIN, HUMIDITY_MAX

    ########################################################################
    # System Universal Boolean Get Methods

    def is_blower_active(self) -> bool:
        """Returns True if the blower is active
        :return: bool.
        """
        return self.get_system_status() not in BLOWER_OFF_STATUSES

    def is_emergency_heat_active(self) -> bool:
        """Returns True if the emergency/aux heat is active
        :return: bool.
        """
        if self.has_emergency_heat():
            return self.get_thermostat_settings_key("emergency_heat")["current_value"]
        raise RuntimeError("This system does not support emergency heat")

    ########################################################################
    # System Universal Get Methods

    def get_fan_modes(self) -> list[str]:
        """Returns the list of fan modes the device supports.

        :return:
        """
        options = self.get_thermostat_settings_key("fan_mode")["options"]
        return [opt["label"] for opt in options]

    def get_fan_mode(self) -> str | None:
        """Returns the current fan mode. See get_fan_modes for the available options.
        :return: str.
        """
        fan_mode = self.get_thermostat_settings_key("fan_mode")
        current_value = fan_mode["current_value"]
        options = fan_mode["options"]
        for opt in options:
            if opt["value"] == current_value:
                return opt["label"]
        return None

    def get_outdoor_temperature(self) -> float | None:
        """Returns the outdoor temperature.
        :return: float - the temperature, returns None if invalid.
        """
        if self.has_outdoor_temperature():
            outdoor_temp = self._get_thermostat_key("outdoor_temperature")
            return float(outdoor_temp) if is_number(outdoor_temp) else None
        raise RuntimeError("This system does not have an outdoor temperature sensor")

    def get_relative_humidity(self) -> float | None:
        """Returns the indoor relative humidity as a percent (0-1)
        :return: float.
        """
        if self.has_relative_humidity():
            try:
                return float(self._get_thermostat_key("indoor_humidity")) / 100
            except ValueError:
                # this has the value "--" when data is unavailable
                return None

        raise RuntimeError("This system does not have a relative humidity sensor.")

    def get_current_compressor_speed(self) -> float:
        """Returns the variable compressor speed, if supported, as a percent (0-1)
        :return: float.
        """
        thermostat_compressor_speed = self._get_thermostat_features_key_or_none(
            "thermostat_compressor_speed",
        )
        if thermostat_compressor_speed is None:
            return 0
        return float(thermostat_compressor_speed["compressor_speed"])

    def get_requested_compressor_speed(self) -> float:
        """Returns the variable compressor's requested speed, if supported, as a
        percent (0-1)
        :return: float.
        """
        # mobile api does not have a requested speed
        return self.get_current_compressor_speed()

    def get_fan_speed_setpoint(self) -> float:
        """Returns the current variable fan speed setpoint from 0-1.
        :return: float.
        """
        if self.has_variable_fan_speed():
            return self.get_thermostat_settings_key("fan_speed")["current_value"]
        raise AttributeError("This system does not have variable fan speed.")

    def get_dehumidify_setpoint(self) -> float:
        """Returns the dehumidify setpoint from 0-1
        :return: float.
        """
        if self.has_dehumidify_support():
            return self.get_thermostat_settings_key("dehumidify")["current_value"]

        raise AttributeError("This system does not support dehumidification")

    def get_humidify_setpoint(self) -> float:
        """Returns the dehumidify setpoint from 0-1
        :return: float.
        """
        if self.has_humidify_support():
            return self.get_thermostat_settings_key("humidify")["current_value"]

        raise AttributeError("This system does not support humidification")

    def get_system_status(self) -> str:
        """Returns the system status such as "System Idle" or "Cooling"
        :return: str.
        """
        return (
            self._get_thermostat_key_or_none("system_status")
            or self._get_thermostat_key_or_none("operating_state")
            or self._get_thermostat_features_key("thermostat")["status"]
        )

    def has_air_cleaner(self) -> bool:
        """Returns if the system has an air cleaner.
        :return: bool.
        """
        return bool(self.get_thermostat_settings_key_or_none("air_cleaner_mode"))

    def get_air_cleaner_mode(self) -> str:
        """Returns the system's air cleaner mode
        :return: str.
        """
        return self.get_thermostat_settings_key("air_cleaner_mode")["current_value"]

    ########################################################################
    # System Universal Set Methods

    async def set_fan_mode(self, fan_mode: str) -> None:
        """Sets the fan mode.
        :param fan_mode: string that must be in self.get_fan_modes()
        :return: None.
        """
        fan_mode_data = self.get_thermostat_settings_key("fan_mode")
        current_fan_mode_value = fan_mode_data["current_value"]
        fan_mode_value: str | None = None
        for opt in fan_mode_data["options"]:
            if opt["label"] == fan_mode:
                fan_mode_value = opt["value"]
                break

        if not fan_mode_value:
            raise KeyError(f"Invalid fan mode {fan_mode} specified")

        # API times out if fan_mode is set to same attribute
        if fan_mode_value != current_fan_mode_value:
            await self._post_and_update_thermostat_json(
                ThermostatEndpoint.FAN_MODE, {"value": fan_mode_value}
            )

    async def set_fan_setpoint(self, fan_setpoint: float) -> None:
        """Sets the fan's setpoint speed as a percent in range. You can see the
        limits by calling Nexia.get_variable_fan_speed_limits()
        :param fan_setpoint: float
        :return: None.
        """
        # This call will get the limits, as well as check if this system has
        # a variable speed fan
        min_speed, max_speed = self.get_variable_fan_speed_limits()

        if min_speed <= fan_setpoint <= max_speed:
            await self._post_and_update_thermostat_json(
                ThermostatEndpoint.FAN_SPEED,
                {"value": fan_setpoint},
            )
        else:
            raise ValueError(
                f"The fan setpoint, {fan_setpoint} is not "
                f"between {min_speed} and {max_speed}.",
            )

    async def set_air_cleaner(self, air_cleaner_mode: str) -> None:
        """Sets the air cleaner mode.
        :param air_cleaner_mode: string that must be in
        AIR_CLEANER_MODES
        :return: None.
        """
        air_cleaner_mode = air_cleaner_mode.lower()
        if air_cleaner_mode in AIR_CLEANER_MODES:
            if air_cleaner_mode != self.get_air_cleaner_mode():
                await self._post_and_update_thermostat_json(
                    ThermostatEndpoint.AIR_CLEANER_MODE,
                    {"value": air_cleaner_mode},
                )
        else:
            raise KeyError("Invalid air cleaner mode specified")

    async def set_follow_schedule(self, follow_schedule: bool) -> None:
        """Enables or disables scheduled operation
        :param follow_schedule: bool - True for follow schedule, False for hold
        current setpoints
        :return: None.
        """
        await self._post_and_update_thermostat_json(
            ThermostatEndpoint.SCHEDULING_ENABLED,
            {"value": "true" if follow_schedule else "false"},
        )

    async def set_emergency_heat(self, emergency_heat_on: bool) -> None:
        """Enables or disables emergency / auxiliary heat.
        :param emergency_heat_on: bool - True for enabled, False for Disabled
        :return: None.
        """
        if self.has_emergency_heat():
            await self._post_and_update_thermostat_json(
                ThermostatEndpoint.EMERGENCY_HEAT,
                {"value": "true" if emergency_heat_on else "false"},
            )
        else:
            raise RuntimeError("This thermostat does not support emergency heat.")

    async def set_humidity_setpoints(  # noqa: C901
        self,
        dehumidify_setpoint: float | None = None,
        humidify_setpoint: float | None = None,
    ) -> None:
        """:param dehumidify_setpoint: float - The dehumidify_setpoint, 0-1, disable: None
        :param humidify_setpoint: float - The humidify setpoint, 0-1, disable: None
        :return:
        """
        if dehumidify_setpoint is None and humidify_setpoint is None:
            # Do nothing
            return

        if not self.has_relative_humidity():
            raise RuntimeError(
                "Setting target humidity is not supported on this thermostat.",
            )

        if self.has_humidify_support():
            humidify_supported = True
            min_humidify, max_humidify = self.get_humidify_setpoint_limits()
            if humidify_setpoint is None:
                humidify_setpoint = self.get_humidify_setpoint()
        else:
            if humidify_setpoint is not None:
                raise RuntimeError("This thermostat does not support humidifying.")
            humidify_supported = False
            humidify_setpoint = 0

        if self.has_dehumidify_support():
            dehumidify_supported = True
            min_dehumidify, max_dehumidify = self.get_dehumidify_setpoint_limits()
            if dehumidify_setpoint is None:
                dehumidify_setpoint = self.get_dehumidify_setpoint()
        else:
            if dehumidify_setpoint is not None:
                raise RuntimeError("This thermostat does not support dehumidifying.")
            dehumidify_supported = False
            dehumidify_setpoint = 0

        # Clean up input
        dehumidify_setpoint = find_humidity_setpoint(dehumidify_setpoint)
        humidify_setpoint = find_humidity_setpoint(humidify_setpoint)

        if (
            dehumidify_supported
            and (
                clamped_dehumidify_setpoint := clamp_to_predefined_values(
                    dehumidify_setpoint,
                    self.dehumidify_setpoints,
                )
            )
            and clamped_dehumidify_setpoint != self.get_dehumidify_setpoint()
        ):
            await self._post_and_update_thermostat_json(
                ThermostatEndpoint.DEHUMIDIFY,
                {"value": str(clamped_dehumidify_setpoint)},
            )

        if (
            humidify_supported
            and (
                clamped_humidify_setpoint := clamp_to_predefined_values(
                    humidify_setpoint,
                    self.humidify_setpoints,
                )
            )
            and clamped_humidify_setpoint != self.get_humidify_setpoint()
        ):
            await self._post_and_update_thermostat_json(
                ThermostatEndpoint.HUMIDIFY,
                {"value": str(clamped_humidify_setpoint)},
            )

    async def set_dehumidify_setpoint(self, dehumidify_setpoint: float) -> None:
        """Sets the overall system's dehumidify setpoint as a percent (0-1).

        The system must support
        :param dehumidify_setpoint: float
        :return: None
        """
        await self.set_humidity_setpoints(dehumidify_setpoint=dehumidify_setpoint)

    async def set_humidify_setpoint(self, humidify_setpoint: float) -> None:
        """Sets the overall system's humidify setpoint as a percent (0-1).

        The system must support
        :param humidify_setpoint: float
        :return: None
        """
        await self.set_humidity_setpoints(humidify_setpoint=humidify_setpoint)

    async def refresh_thermostat_data(self) -> None:
        """Refresh data in this thermostat instance.
        Note: Many other methods refresh this data before completing.
        :return: None
        """
        self_ref = f"{self._nexia_home.mobile_url}/xxl_thermostats/{self.thermostat_id}"

        async with await self._nexia_home._get_url(self_ref) as response:  # noqa: SLF001
            self.update_thermostat_json((await response.json())["result"])

    ########################################################################
    # Zone Get Methods

    def get_zone_ids(self) -> list[int]:
        """Returns a list of available zone IDs with a starting index of 0.
        :return: list(int).
        """
        # The zones are in a list, so there are no keys to pull out. I have to
        # create a new list of IDs.
        return [zone.zone_id for zone in self.zones]

    def get_zone_by_id(self, zone_id: int) -> NexiaThermostatZone:
        """Get a zone by its nexia id."""
        for zone in self.zones:
            if zone.zone_id == zone_id:
                return zone
        valid_ids = (str(id_) for id_ in self.get_zone_ids())
        raise KeyError(f"Zone ID not found, valid IDs are: {', '.join(valid_ids)}")

    def _get_thermostat_deep_key(
        self,
        area: str,
        area_primary_key: str,
        key: str,
    ) -> Any:
        """Returns the thermostat value from deep inside the thermostat's
        JSON.
        :param area: The area of the json to look i.e. "settings", "features", etc.
        :param area_primary_key: The name of the primary key such as "name" or "key"
        :param key: str
        :return: value.
        """
        data = find_dict_with_keyvalue_in_json(
            self._thermostat_json[area],
            area_primary_key,
            key,
        )

        if not data:
            raise KeyError(f'Key "{key}" not in the thermostat JSON!')
        return data

    def _get_thermostat_features_key_or_none(self, key: str):
        """Returns the thermostat value from the provided key in the thermostat's
        JSON.
        :param key: str
        :return: value.
        """
        try:
            return self._get_thermostat_features_key(key)
        except KeyError:
            return None

    def _get_thermostat_features_key(self, key: str):
        """Returns the thermostat value from the provided key in the thermostat's
        JSON.
        :param key: str
        :return: value.
        """
        return self._get_thermostat_deep_key("features", "name", key)

    def _get_thermostat_key_or_none(self, key):
        """Returns the thermostat value from the provided key in the thermostat's
        JSON.
        :param key: str
        :return: value.
        """
        return self._thermostat_json.get(key)

    def _get_thermostat_key(self, key):
        """Returns the thermostat value from the provided key in the thermostat's
        JSON.
        :param key: str
        :return: value.
        """
        thermostat = self._thermostat_json
        if key in thermostat:
            return thermostat[key]

        raise KeyError(f'Key "{key}" not in the thermostat JSON ({thermostat}!')

    def get_thermostat_settings_key_or_none(self, key):
        """Returns the thermostat value from the provided key in the thermostat's
        JSON.
        :param key: str
        :return: value.
        """
        try:
            return self.get_thermostat_settings_key(key)
        except KeyError:
            return None

    def get_thermostat_settings_key(self, key: str) -> dict[str, Any]:
        """Returns the thermostat value from the provided key in the thermostat's
        JSON.
        :param key: str
        :return: value.
        """
        return self._get_thermostat_deep_key("settings", "type", key)

    def _get_zone_json(self, zone_id=0):
        """Returns the thermostat zone's JSON
        :param zone_id: The index of the zone, defaults to 0.
        :return: dict(thermostat_json['zones'][zone_id]).
        """
        thermostat = self._thermostat_json
        if not thermostat:
            return None

        zone = find_dict_with_keyvalue_in_json(thermostat["zones"], "id", zone_id)

        if not zone:
            raise IndexError(
                f"The zone_id ({zone_id}) does not exist in the thermostat zones.",
            )
        return zone

    async def _post_and_update_thermostat_json(
        self, end_point: ThermostatEndpoint, payload: dict[str, Any]
    ) -> None:
        """Post to the thermostat and update the thermostat json."""
        if not (end_point_data := ENDPOINT_MAP.get(end_point)):
            raise ValueError(f"Invalid endpoint {end_point}")

        url: str | None = None
        method: str = DEFAULT_UPDATE_METHOD
        try:
            actions: dict[str, dict[str, str]] = self._get_thermostat_deep_key(
                end_point_data.area,
                end_point_data.area_primary_key,
                end_point_data.key,
            )["actions"]
        except KeyError:
            pass
        else:
            for action in ("self", end_point_data.action):
                if action_data := actions.get(action):
                    url = action_data["href"]
                    method = action_data.get("method", DEFAULT_UPDATE_METHOD)
                    break

        if url is None:
            url = self.API_MOBILE_THERMOSTAT_URL.format(
                end_point=end_point_data.fallback_endpoint,
                thermostat_id=self.thermostat_id,
            )

        if method != DEFAULT_UPDATE_METHOD:
            raise ValueError(
                f"Unsupported method {method} for endpoint {end_point} url {url}"
            )

        async with await self._nexia_home.post_url(url, payload) as response:
            self.update_thermostat_json((await response.json())["result"])

    def update_thermostat_json(self, thermostat_json):
        """Update with new json from the api."""
        if self._thermostat_json is None:
            return

        _LOGGER.debug(
            "Updated thermostat_id:%s with new data from post",
            self.thermostat_id,
        )
        self._thermostat_json.update(thermostat_json)

        zone_updates_by_id = {}
        for zone_json in thermostat_json["zones"]:
            zone_updates_by_id[zone_json["id"]] = zone_json

        for zone in self.zones:
            if zone.zone_id in zone_updates_by_id:
                zone.update_zone_json(zone_updates_by_id[zone.zone_id])
